from .test import (
    ValidationTest,
    ComplianceLevel,
    TestCategory,
    IIIFVersion,
    TargetServer,
    ValidationFailure,
    ValidationSuccess,
    ImageAPIRequest,
    get_image_format,
)


class ImageReturned(ValidationTest):
    name = "Image is returned"
    compliance_level = ComplianceLevel.LEVEL_0
    category = TestCategory.FORMAT
    versions = [IIIFVersion.V2, IIIFVersion.V3]

    @staticmethod
    def run(server: TargetServer) -> ValidationFailure | ValidationSuccess:
        req = ImageAPIRequest.of()
        try:
            get_image_format(server, req)
            return ValidationSuccess(details="Successfully received full image")
        except Exception as e:
            url = req.url(server)
            return ValidationFailure(
                url=url,
                expected="Full JPEG image",
                received=f"Invalid image due to error: {e}",
                details="Server did not return an image for the validation identifier",
            )



from .test import (
    ValidationTest,
    ComplianceLevel,
    TestCategory,
    IIIFVersion,
    TargetServer,
    ValidationFailure,
    ValidationSuccess,
    make_request,
)


class ErrorUnescapedIdentifier(ValidationTest):
    name = "Unescaped Identifier gives 404"
    compliance_level = ComplianceLevel.LEVEL_1
    category = TestCategory.INFO
    versions = [IIIFVersion.V2, IIIFVersion.V3]

    @staticmethod
    def run(server: TargetServer) -> ValidationFailure | ValidationSuccess:
        url = f"{server.base_url}/%5Bfrob%5D/info.json"
        resp = make_request(url)
        if resp.status in (404, 400):
            return ValidationSuccess(
                details="Returned 404 or 400 for unescaped identifier"
            )
        else:
            return ValidationFailure(
                url=url,
                expected="HTTP Status 404 or 400",
                received=f"HTTP Status {resp.status}",
                details="Got unexpected status code for unescaped identifier",
            )



import random
from .test import (
    ValidationTest,
    ComplianceLevel,
    TestCategory,
    IIIFVersion,
    TargetServer,
    ValidationFailure,
    ValidationSuccess,
    ImageAPIRequest,
    make_request,
    make_random_string,
)


class SizeErrorRandomTest(ValidationTest):
    name = "Random size gives 400"
    compliance_level = ComplianceLevel.LEVEL_1
    category = TestCategory.SIZE
    versions = [IIIFVersion.V2, IIIFVersion.V3]

    @staticmethod
    def run(server: TargetServer) -> list[ValidationSuccess | ValidationFailure]:
        random.seed(31337)
        results = []
        for i in range(4):
            random_size = make_random_string(6)
            req = ImageAPIRequest.of(size=random_size)
            url = req.url(server)
            resp = make_request(url)
            if resp.status == 400:
                results.append(
                    ValidationSuccess(
                        details=f"Received 400 for random size='{random_size}'"
                    )
                )
            else:
                results.append(
                    ValidationFailure(
                        url=url,
                        expected="400",
                        received=f"{resp.status}",
                        details=f"Did not receive 400 for random size='{random_size}'",
                    )
                )
        return results



import random

from .test import (
    ValidationTest,
    TargetServer,
    ComplianceLevel,
    TestCategory,
    IIIFVersion,
    ValidationFailure,
    ValidationSuccess,
    ImageAPIRequest,
    make_request,
    make_random_string,
)


class TestFormatErrorRandom(ValidationTest):
    name = "Random format gives 400"
    compliance_level = ComplianceLevel.LEVEL_1
    category = TestCategory.FORMAT
    versions = [IIIFVersion.V2, IIIFVersion.V3]

    @staticmethod
    def run(server: TargetServer) -> ValidationFailure | ValidationSuccess:
        random.seed(31337)
        for _ in range(16):
            random_format = make_random_string(3)
            url = ImageAPIRequest.of(format=random_format).url(server)
            resp = make_request(url)
            if resp.status == 400:
                continue  # This is the expected outcome
            return ValidationFailure(
                url=url,
                expected="400 Bad Request",
                received=f"{resp.status} {resp.body.decode('utf8')}",
                details="Server did not return 400 for random format",
            )
        return ValidationSuccess(details="Server returned 400 for random formats")



import random

from .test import (
    ValidationTest,
    ComplianceLevel,
    TestCategory,
    IIIFVersion,
    TargetServer,
    ValidationFailure,
    ValidationSuccess,
    ImageAPIRequest,
    get_image,
    get_expected_image,
    compare_images,
)


class SizePercentTest(ValidationTest):
    name = "Size specified by percent"
    compliance_level = {
        IIIFVersion.V3: ComplianceLevel.LEVEL_2,
        IIIFVersion.V2: ComplianceLevel.LEVEL_1,
    }
    category = TestCategory.SIZE
    versions = [IIIFVersion.V2, IIIFVersion.V3]
    extra_name = "sizeByPct"

    @staticmethod
    def run(server: TargetServer) -> list[ValidationSuccess | ValidationFailure]:
        random.seed(31337)
        results = []
        full_expected_img = get_expected_image()
        for i in range(4):
            s = random.randint(45, 75)
            req = ImageAPIRequest.of(size=f"pct:{s}")
            img = get_image(server, req)

            expected_size = (s * 10, s * 10)
            if (img.width, img.height) != expected_size:
                url = req.url(server)
                results.append(
                    ValidationFailure(
                        url=url,
                        expected=f"Image size to be {expected_size}",
                        received=f"{img.size}",
                        details=f"Incorrect image size for size='pct:{s}'",
                    )
                )
                continue

            expected_img = full_expected_img.thumbnail_image(
                s * 10, height=s * 10, size="force"
            )
            if not compare_images(img, expected_img):
                url = req.url(server)
                results.append(
                    ValidationFailure(
                        url=url,
                        expected="Image content to match validation image",
                        received="Different image content",
                        details=f"Image content incorrect for size='pct:{s}'",
                    )
                )
                continue

            results.append(
                ValidationSuccess(
                    details=f"Image size and content correct for size='pct:{s}'"
                )
            )
        return results



from .test import (
    ValidationTest,
    ComplianceLevel,
    TestCategory,
    IIIFVersion,
    TargetServer,
    ValidationFailure,
    ValidationSuccess,
    ImageAPIRequest,
    get_image,
    get_expected_image,
    compare_images,
)


class RotationFullBasicTest(ValidationTest):
    name = "Rotation by 90 degree values"
    compliance_level = ComplianceLevel.LEVEL_2
    category = TestCategory.ROTATION
    versions = [IIIFVersion.V2, IIIFVersion.V3]
    extra_name = "rotationBy90s"

    @staticmethod
    def run(server: TargetServer) -> list[ValidationSuccess | ValidationFailure]:
        results = []
        expected_img = get_expected_image()
        # 0 degree rotation is implicitly tested elsewhere
        rotations = [90, 180, 270]
        for rot in rotations:
            request = ImageAPIRequest.of(rotation=str(rot))
            try:
                rotated_img = get_image(server, request)
                if rot == 180:
                    expected_rotated_img = expected_img.rot180()
                elif rot == 90:
                    expected_rotated_img = expected_img.rot90()
                else:
                    expected_rotated_img = expected_img.rot270()

                if compare_images(rotated_img, expected_rotated_img):
                    results.append(
                        ValidationSuccess(
                            details=f"Rotation by {rot} degrees was successful."
                        )
                    )
                else:
                    url = request.url(server)
                    results.append(
                        ValidationFailure(
                            url=url,
                            expected="A correctly rotated image.",
                            received="An incorrectly rotated image.",
                            details=f"Image rotated by {rot} degrees does not match the expected image.",
                        )
                    )
            except Exception as e:
                url = request.url(server)
                results.append(
                    ValidationFailure(
                        url=url,
                        expected="A valid image.",
                        received=f"An error: {e}",
                        details=f"Request for rotation {rot} failed.",
                    )
                )
        return results



from typing import Literal
from urllib.request import Request

from .test import (
    ValidationTest,
    ComplianceLevel,
    TestCategory,
    IIIFVersion,
    TargetServer,
    ValidationFailure,
    ValidationSuccess,
    make_request,
    HEADERS,
    ImageAPIRequest,
)


class CorsImageTest(ValidationTest):
    """
    [...]
    Servers SHOULD support CORS on image responses.
    [...]
    Servers SHOULD support reuse of Image API resources by following the
    relevant requirements of the CORS specification, including the
    `Access-Control-Allow-Origin` header and the preflight request pattern.
    """

    name = "CORS Header on all responses"
    compliance_level = ComplianceLevel.LEVEL_1
    category = TestCategory.HTTP
    versions = [IIIFVersion.V2, IIIFVersion.V3]
    extra_name = "cors"

    @staticmethod
    def run(server: TargetServer) -> list[ValidationSuccess | ValidationFailure]:
        results = []
        for resource_type in ("info", "image"):
            for method in ("OPTIONS", "GET"):
                results.append(_single_run(server, resource_type, method))
        return results


def _single_run(
    server: TargetServer,
    resource_type: Literal["info", "image"],
    method: Literal["GET", "OPTIONS"],
) -> ValidationSuccess | ValidationFailure:
    origin = HEADERS["Origin"]

    if resource_type == "info":
        url = server.info_url()
    else:
        url = ImageAPIRequest.of().url(server)
    resp = make_request(Request(url, method=method))
    cors = resp.headers.get("access-control-allow-origin", None)
    if cors in ("*", origin):
        return ValidationSuccess(
            details=f"CORS header on {resource_type} with method={method} present and correct"
        )
    else:
        return ValidationFailure(
            url=url,
            expected=f"Access-Control-Allow-Origin: {origin} or *",
            received=f"Access-Control-Allow-Origin: {cors or '<none>'}",
            details=f"CORS header for {resource_type} with method={method} missing or incorrect",
        )



import random
import uuid

from .test import (
    ValidationTest,
    ComplianceLevel,
    TestCategory,
    IIIFVersion,
    TargetServer,
    ValidationFailure,
    ValidationSuccess,
    make_request,
)


class ErrorRandomId(ValidationTest):
    name = "Random Identifier gives 404"
    compliance_level = ComplianceLevel.LEVEL_1
    category = TestCategory.INFO
    versions = [IIIFVersion.V2, IIIFVersion.V3]

    @staticmethod
    def run(server: TargetServer) -> ValidationFailure | ValidationSuccess:
        # For reproducability
        random.seed(31337)
        random_id = uuid.uuid1()
        url = f"{server.base_url}/{random_id}/info.json"
        resp = make_request(url)
        if resp.status == 404:
            return ValidationSuccess(details="Returned 404 for random identifier")
        else:
            return ValidationFailure(
                url=url,
                expected="HTTP Status 404",
                received=f"HTTP Status {resp.status}",
                details="Got unexpected status code for random identifier",
            )



import random

from .test import (
    ValidationTest,
    ComplianceLevel,
    TestCategory,
    IIIFVersion,
    TargetServer,
    ValidationFailure,
    ValidationSuccess,
    ImageAPIRequest,
    get_image,
    get_expected_image,
    compare_images,
)


class RotationRegionBasicTest(ValidationTest):
    name = "Rotation of region by 90 degree values"
    compliance_level = ComplianceLevel.LEVEL_2
    category = TestCategory.ROTATION
    versions = [IIIFVersion.V2, IIIFVersion.V3]
    extra_name = "rotationBy90s"

    @staticmethod
    def run(server: TargetServer) -> list[ValidationSuccess | ValidationFailure]:
        results = []
        expected_img = get_expected_image()
        rotations = [90, 180, 270]

        for i in range(2):
            x = random.randint(0, 500)
            y = random.randint(0, 500)
            w = random.randint(100, 400)
            h = random.randint(100, 400)
            region_str = f"{x},{y},{w},{h}"

            for rot in rotations:
                request = ImageAPIRequest.of(region=region_str, rotation=str(rot))
                img_from_server = get_image(server, request)

                expected_region = expected_img.extract_area(x, y, w, h)
                if rot == 90:
                    expected_rotated_region = expected_region.rot90()
                elif rot == 180:
                    expected_rotated_region = expected_region.rot180()
                else:
                    expected_rotated_region = expected_region.rot270()

                expected_w, expected_h = (
                    expected_rotated_region.width,
                    expected_rotated_region.height,
                )
                if (
                    img_from_server.width != expected_w
                    or img_from_server.height != expected_h
                ):
                    url = request.url(server)
                    results.append(
                        ValidationFailure(
                            url=url,
                            expected=f"Image dimensions to be ({expected_w}, {expected_h})",
                            received=f"Image dimensions were ({img_from_server.width}, {img_from_server.height})",
                            details="Incorrect image dimensions for rotated region.",
                        )
                    )
                    continue

                if compare_images(img_from_server, expected_rotated_region):
                    results.append(
                        ValidationSuccess(
                            details=f"Rotation of region '{region_str}' by {rot} degrees was successful."
                        )
                    )
                else:
                    url = request.url(server)
                    results.append(
                        ValidationFailure(
                            url=url,
                            expected="A correctly rotated region.",
                            received="An incorrectly rotated region.",
                            details=f"Image region {region_str} rotated by {rot} degrees does not match the expected image.",
                        )
                    )
        return results



from .test import (
    ValidationTest,
    ComplianceLevel,
    TestCategory,
    IIIFVersion,
    TargetServer,
    ValidationFailure,
    ValidationSuccess,
    make_request,
)


class InfoHasJsonLdTypeWhenRequested(ValidationTest):
    name = "Info response has JSON-LD type when requested"
    compliance_level = ComplianceLevel.LEVEL_1
    category = TestCategory.HTTP
    versions = [IIIFVersion.V2, IIIFVersion.V3]
    extra_name = "jsonldMediaType"

    @staticmethod
    def run(server: TargetServer) -> ValidationFailure | ValidationSuccess:
        info_resp = make_request(server.info_url(), {"Accept": "application/ld+json"})
        if info_resp.status != 200:
            url = server.info_url()
            return ValidationFailure(
                url=url,
                expected="HTTP Status 200",
                received=f"HTTP Status {info_resp.status}",
                details="Got unexpected status code when requesting JSON-LD",
            )
        content_type = info_resp.headers.get("content-type", "")
        if not content_type.startswith("application/ld+json"):
            url = server.info_url()
            return ValidationFailure(
                url=url,
                expected="Content-Type to start with application/ld+json",
                received=f"Content-Type: {content_type}",
                details="Got unexpected Content-Type when requesting JSON-LD",
            )
        return ValidationSuccess(
            details="Info response had JSON-LD type when requested"
        )



from typing import Literal
from urllib.request import Request

from .test import (
    ValidationTest,
    ComplianceLevel,
    TestCategory,
    IIIFVersion,
    TargetServer,
    ValidationFailure,
    ValidationSuccess,
    make_request,
    ImageAPIRequest,
)


class HttpCachingTest(ValidationTest):
    name = "Proper HTTP Caching implementation for info.json and images"
    compliance_level = ComplianceLevel.OPTIONAL
    category = TestCategory.HTTP
    versions = [IIIFVersion.V2, IIIFVersion.V3]

    @staticmethod
    def run(server: TargetServer) -> list[ValidationSuccess | ValidationFailure]:
        results = []
        for resource_type in ("info", "image"):
            results.extend(_single_run(server, resource_type))
        return results


def _single_run(
    server: TargetServer,
    resource_type: Literal["info", "image"],
) -> list[ValidationSuccess | ValidationFailure]:
    results = []
    etag = None
    last_modified = None
    if resource_type == "info":
        url = server.info_url()
    else:
        url = ImageAPIRequest.of().url(server)
    for method in ("HEAD", "GET"):
        resp = make_request(Request(url, method=method))
        etag = resp.headers.get("etag", None)
        if etag is not None:
            results.append(
                ValidationSuccess(
                    details=f"ETag header on {resource_type} with method={method} present"
                )
            )
            continue
        last_modified = resp.headers.get("last-modified", None)
        if last_modified is not None:
            results.append(
                ValidationSuccess(
                    details=f"Last-Modified header on {resource_type} with method={method} present"
                )
            )
            continue
        results.append(
            ValidationFailure(
                url=url,
                expected="Either ETag or Last-Modified header",
                received="Neither ETag nor Last-Modified header",
                details=f"Server did not provide any caching headers on {resource_type} with method={method}",
            )
        )
    extra_headers = {}
    if etag is not None:
        extra_headers["If-None-Match"] = etag
    if last_modified is not None:
        extra_headers["If-Modified-Since"] = last_modified
    else:
        return results

    resp = make_request(url, extra_headers=extra_headers)
    if resp.status == 304:
        results.append(
            ValidationSuccess(
                details=f"Server returned HTTP 304 Not Modified for a conditional request on the validation {resource_type}"
            )
        )
    else:
        results.append(
            ValidationFailure(
                url=url,
                expected="304 Not Modified",
                received=f"{resp.status} {resp.body.decode('utf-8', errors='replace')}",
                details=f"Server did not return the expected response on the validation {resource_type}.",
            )
        )
    return results



import random

from .test import (
    ValidationTest,
    ComplianceLevel,
    TestCategory,
    IIIFVersion,
    TargetServer,
    ValidationFailure,
    ValidationSuccess,
    ImageAPIRequest,
    get_expected_image,
    get_image,
    compare_images,
)


class RegionSquare(ValidationTest):
    name = "Request a square region of the full image."
    compliance_level = {
        IIIFVersion.V3: ComplianceLevel.LEVEL_1,
        IIIFVersion.V2: ComplianceLevel.OPTIONAL,
    }
    category = TestCategory.REGION
    versions = [IIIFVersion.V2, IIIFVersion.V3]
    extra_name = "regionSquare"

    @staticmethod
    def run(server: TargetServer) -> ValidationFailure | ValidationSuccess:
        random.seed(31337)  # for reproducibility
        # Square region on full image should return the full validation image, since its width/height are equal
        expected = get_expected_image()
        req = ImageAPIRequest.of(region="square")
        square = get_image(server, req)
        if not compare_images(expected, square):
            url = req.url(server)
            return ValidationFailure(
                url=url,
                expected="Square region to match expected image",
                received="Square region did not match expected image",
                details="Server did not correctly handle region=square",
            )
        return ValidationSuccess(details="Server correctly handled region=square")



import random

from .test import (
    ValidationTest,
    ComplianceLevel,
    TestCategory,
    IIIFVersion,
    TargetServer,
    ValidationFailure,
    ValidationSuccess,
    ImageAPIRequest,
    get_image,
    get_expected_image,
    compare_images,
    make_request,
)


class SizeUpTest(ValidationTest):
    name = "Size greater than 100%"
    compliance_level = ComplianceLevel.OPTIONAL
    category = TestCategory.SIZE
    versions = [IIIFVersion.V2, IIIFVersion.V3]
    extra_name = "sizeUpscaling"

    @staticmethod
    def run(server: TargetServer) -> list[ValidationSuccess | ValidationFailure]:
        random.seed(31337)
        if server.version == IIIFVersion.V3:
            return SizeUpTest.run_v3(server)
        else:
            return SizeUpTest.run_v2(server)

    @staticmethod
    def run_v2(server: TargetServer) -> list[ValidationSuccess | ValidationFailure]:
        s = random.randint(1100, 2000)
        req = ImageAPIRequest.of(size=f",{s}")
        img = get_image(server, req)

        expected_size = (s, s)
        if (img.width, img.height) != expected_size:
            url = req.url(server)
            return [
                ValidationFailure(
                    url=url,
                    expected=f"Image size to be {expected_size}",
                    received=f"{img.size}",
                    details=f"Incorrect image size for upscaled size=',{s}'",
                )
            ]

        full_expected_img = get_expected_image()
        expected_img = full_expected_img.thumbnail_image(s, height=s, size="force")
        if not compare_images(img, expected_img):
            url = req.url(server)
            return [
                ValidationFailure(
                    url=url,
                    expected="Image content to match validation image",
                    received="Different image content",
                    details=f"Image content incorrect for upscaled size=',{s}'",
                )
            ]

        return [
            ValidationSuccess(
                details=f"Image size and content correct for upscaled size=',{s}'"
            )
        ]

    @staticmethod
    def run_v3(server: TargetServer) -> list[ValidationSuccess | ValidationFailure]:
        results = []
        s = random.randint(1100, 2000)

        # Test that upscaling without ^ fails
        req = ImageAPIRequest.of(size=f",{s}")
        url = req.url(server)
        resp = make_request(url)
        if resp.status != 200:
            results.append(
                ValidationSuccess(
                    details=f"Upscaling size=',{s}' failed as expected in v3 without '^'."
                )
            )
        else:
            results.append(
                ValidationFailure(
                    url=url,
                    expected="Non-200 status code",
                    received=f"{resp.status}",
                    details=f"Upscaling size=',{s}' should fail in v3 without '^'.",
                )
            )

        # Test that upscaling with ^ works
        sizes_to_check = {
            f"^{s},{s}": (s, s),
            f"^,{s}": (s, s),
            f"^{s},": (s, s),
            "^max": (1000, 1000),
            "^pct:200": (2000, 2000),
            "^!2000,500": (500, 500),
            "^!2000,3000": (2000, 2000),
        }

        full_expected_img = get_expected_image()
        for size_str, expected_size in sizes_to_check.items():
            req = ImageAPIRequest.of(size=size_str)
            img = get_image(server, req)

            if (img.width, img.height) != expected_size:
                url = req.url(server)
                results.append(
                    ValidationFailure(
                        url=url,
                        expected=f"Image size to be {expected_size}",
                        received=f"{img.size}",
                        details=f"Incorrect image size for upscaled size='{size_str}'",
                    )
                )
                continue

            expected_img = full_expected_img.thumbnail_image(
                expected_size[0], height=expected_size[1], size="force"
            )
            if not compare_images(img, expected_img):
                url = req.url(server)
                results.append(
                    ValidationFailure(
                        url=url,
                        expected="Image content to match validation image",
                        received="Different image content",
                        details=f"Image content incorrect for upscaled size='{size_str}'",
                    )
                )
                continue

            results.append(
                ValidationSuccess(
                    details=f"Image size and content correct for upscaled size='{size_str}'"
                )
            )

        return results



import random

from .test import (
    ValidationTest,
    ComplianceLevel,
    TestCategory,
    IIIFVersion,
    TargetServer,
    ValidationFailure,
    ValidationSuccess,
    ImageAPIRequest,
    get_image,
    get_expected_image,
    compare_images,
)


class RotationFullNon90Test(ValidationTest):
    name = "Rotation by non 90 degree values"
    compliance_level = ComplianceLevel.OPTIONAL
    category = TestCategory.ROTATION
    versions = [IIIFVersion.V2, IIIFVersion.V3]
    extra_name = "rotationArbitrary"

    @staticmethod
    def run(server: TargetServer) -> ValidationSuccess | ValidationFailure:
        # Pick a random rotation that isn't a multiple of 90
        rot = random.randint(1, 359)
        while rot % 90 == 0:
            rot = random.randint(1, 359)

        expected_image = get_expected_image().rotate(rot)
        request = ImageAPIRequest.of(rotation=str(rot))
        actual_image = get_image(server, request)
        if not compare_images(
            actual_image.extract_area(300, 300, 300, 300),
            expected_image.extract_area(300, 300, 300, 300),
        ):
            url = request.url(server)
            return ValidationFailure(
                url=url,
                expected="A correctly rotated image.",
                received="An incorrectly rotated image.",
                details=f"Image rotated by {rot} degrees does not match the expected image.",
            )
        return ValidationSuccess(
            details=f"Server successfully returned an image for a non-90 degree rotation ({rot})."
        )



import argparse

from collections import defaultdict

import rich
from rich.progress import Progress, BarColumn

from .validate import run_tests
from .tests.test import TargetServer, IIIFVersion, ValidationFailure, ValidationTest


def main(
    base_url: str,
    is_v2: bool,
    validation_id="67352ccc-d1b0-11e1-89ae-279075081939",
    max_threads: int = 4,
):
    server = TargetServer(
        base_url,
        validation_id=validation_id,
        version=IIIFVersion.V2 if is_v2 else IIIFVersion.V3,
    )

    rich.print(
        f"Validating IIIF Image API {server.version.value} server at [blue]{base_url}[/blue]"
    )
    features = server.feature_set()
    rich.print(
        f"Detected compliance level: [green]{features.compliance_level.name}[/green]"
    )
    if features.extra_qualities:
        rich.print(
            f"Detected extra qualities: [green]{', '.join(features.extra_qualities)}[/green]"
        )
    if features.extra_formats:
        rich.print(
            f"Detected extra formats: [green]{', '.join(features.extra_formats)}[/green]"
        )
    if features.extra_features:
        rich.print(
            f"Detected extra features: [green]{', '.join(features.extra_features)}[/green]"
        )

    test_gen = run_tests(server, max_threads=max_threads)

    failures: dict[type[ValidationTest], list[ValidationFailure]] = defaultdict(list)
    with Progress(
        "[progress.description]{task.description}",
        BarColumn(),
        "[progress.percentage]{task.percentage:>3.0f}%",
        "[green]✓ {task.fields[ok]}[/green]",
        "[red]✗ {task.fields[failed]}[/red]",
        "[blue]{task.fields[passed_percent]:<2.0f}% passed[/blue]",
    ) as progress:
        failed_count = 0
        ok_count = 0
        task = progress.add_task(
            "Running tests...",
            total=test_gen.num_total,
            ok=0,
            failed=0,
            passed_percent=0.0,
        )
        for test, result in test_gen:
            if not isinstance(result, list):
                result = [result]
            for res in result:
                if isinstance(res, ValidationFailure):
                    failed_count += 1
                    failures[test].append(res)
                    iiif_name = (
                        f"[blue]({str(test.extra_name)})[/blue] "
                        if test.extra_name
                        else ""
                    )
                    progress.console.print(
                        f"[red]✗ {test.name} {iiif_name}[white]\\[{test.__module__.split('.')[-1]}][/white]: {res.details}[/red]\n"
                        f"  Expected: [blue]{res.expected}[/blue]\n"
                        f"  Received: [red]{res.received}[/red]\n"
                        f"  URL: {res.url}\n"
                    )
                else:
                    ok_count += 1
            progress.update(
                task,
                advance=1,
                failed=failed_count,
                ok=ok_count,
                passed_percent=ok_count / max((failed_count + ok_count), 1) * 100,
            )


def cli():
    parser = argparse.ArgumentParser(
        description="Validate an IIIF Image API 2.0 or 3.0 server."
    )
    parser.add_argument("base_url", type=str, help="The base URL of the IIIF server.")
    parser.add_argument(
        "--v2",
        action="store_true",
        help="Validate as an IIIF Image API 2.0 server (default is 3.0).",
    )
    parser.add_argument(
        "--validation-id",
        type=str,
        default="67352ccc-d1b0-11e1-89ae-279075081939",
        help="The validation ID to use (default is the official IIIF validator ID).",
    )
    parser.add_argument(
        "--max-threads",
        type=int,
        default=4,
        help="The maximum number of threads to use for running the tests (default is 4).",
    )
    args = parser.parse_args()

    main(
        args.base_url,
        args.v2,
        validation_id=args.validation_id,
        max_threads=args.max_threads,
    )


if __name__ == "__main__":
    cli()



import random

from .test import (
    ValidationTest,
    ComplianceLevel,
    TestCategory,
    IIIFVersion,
    TargetServer,
    ValidationFailure,
    ValidationSuccess,
    ImageAPIRequest,
    get_image,
    get_expected_image,
    compare_images,
)


class SizeRegionTest(ValidationTest):
    name = "Region at specified size"
    compliance_level = ComplianceLevel.LEVEL_2
    category = TestCategory.SIZE
    versions = [IIIFVersion.V2, IIIFVersion.V3]

    @staticmethod
    def run(server: TargetServer) -> list[ValidationSuccess | ValidationFailure]:
        random.seed(31337)
        full_expected_img = get_expected_image()
        results = []

        for i in range(5):
            s = random.randint(35, 90)
            x = random.randint(0, 9)
            y = random.randint(0, 9)

            region = f"{x * 100},{y * 100},100,100"
            size = f"{s},{s}"
            req = ImageAPIRequest.of(region=region, size=size)
            img = get_image(server, req)

            if img.width != s or img.height != s:
                url = req.url(server)
                results.append(
                    ValidationFailure(
                        url=url,
                        expected=f"({s}, {s})",
                        received=f"({img.width}, {img.height})",
                        details=f"Incorrect image size for region='{region}' and size='{size}'",
                    )
                )
                continue

            expected_region = full_expected_img.extract_area(x * 100, y * 100, 100, 100)
            expected_img = expected_region.thumbnail_image(s, height=s, size="force")

            if compare_images(img, expected_img):
                results.append(
                    ValidationSuccess(
                        details=f"Image content correct for region='{region}' and size='{size}'"
                    )
                )
            else:
                url = req.url(server)
                results.append(
                    ValidationFailure(
                        url=url,
                        expected=f"Image content to match validation image for region {x},{y}",
                        received="Different image content",
                        details=f"Image content incorrect for region='{region}' and size='{size}'",
                    )
                )
        return results



import random

from .test import (
    ValidationTest,
    ComplianceLevel,
    TestCategory,
    IIIFVersion,
    TargetServer,
    ValidationFailure,
    ValidationSuccess,
    ImageAPIRequest,
    get_image,
    get_expected_image,
    compare_images,
)


class RotationRegionNon90Test(ValidationTest):
    name = "Rotation of region by non 90 degree values"
    compliance_level = ComplianceLevel.OPTIONAL
    category = TestCategory.ROTATION
    versions = [IIIFVersion.V2, IIIFVersion.V3]
    extra_name = "rotationArbitrary"

    @staticmethod
    def run(server: TargetServer) -> list[ValidationSuccess | ValidationFailure]:
        results = []
        for i in range(2):
            rot = random.randint(1, 359)
            while rot % 90 == 0:
                rot = random.randint(1, 359)

            x = random.randint(0, 500)
            y = random.randint(0, 500)
            w = random.randint(100, 400)
            h = random.randint(100, 400)
            region_str = f"{x},{y},{w},{h}"

            request = ImageAPIRequest.of(region=region_str, rotation=str(rot))
            img_from_server = get_image(server, request)
            expected_rotated = get_expected_image().extract_area(x, y, w, h).rotate(rot)
            center_server = img_from_server.extract_area(w // 4, h // 4, w // 4, h // 4)
            center_expected = expected_rotated.extract_area(
                w // 4, h // 4, w // 4, h // 4
            )
            if not compare_images(center_server, center_expected):
                url = request.url(server)
                results.append(
                    ValidationFailure(
                        url=url,
                        expected="A correctly rotated region.",
                        received="An incorrectly rotated region.",
                        details=f"Image region {region_str} rotated by {rot} degrees does not match the expected image.",
                    )
                )
            else:
                results.append(
                    ValidationSuccess(
                        details=f"Server successfully returned an image for a non-90 degree rotation ({rot}) of region '{region_str}'."
                    )
                )
        return results



from urllib.request import HTTPRedirectHandler, build_opener

from .test import (
    ValidationTest,
    ComplianceLevel,
    TestCategory,
    IIIFVersion,
    TargetServer,
    ValidationFailure,
    ValidationSuccess,
    make_request,
)


class NoRedirect(HTTPRedirectHandler):
    def redirect_request(self, req, fp, code, msg, headers, newurl):
        # Prevent redirection
        return None


opener = build_opener(NoRedirect)


class BaseUrlRedirect(ValidationTest):
    name = "Base URL Redirects"
    compliance_level = ComplianceLevel.LEVEL_1
    category = TestCategory.HTTP
    versions = [IIIFVersion.V2, IIIFVersion.V3]
    extra_name = "baseUriRedirect"

    @staticmethod
    def run(server: TargetServer) -> ValidationSuccess | ValidationFailure:
        url = f"{server.base_url}/{server.validation_id}"
        expected = server.info_url()
        response = make_request(url, opener=opener)
        if response.status >= 300 and response.status < 400:
            new_url = response.headers.get("location", "")
            if new_url == expected:
                return ValidationSuccess(details=f"Redirected to {new_url}")
            else:
                return ValidationFailure(
                    url=url,
                    expected=f"Redirect to {expected}",
                    received=f"Redirect to {new_url}",
                    details="Invalid redirect location",
                )
        else:
            return ValidationFailure(
                url=url,
                expected="Redirect Status (3xx)",
                received=f"Status {response.status}",
                details="Did not receive a redirect status code",
            )



from .test import (
    ValidationTest,
    ComplianceLevel,
    TestCategory,
    IIIFVersion,
    TargetServer,
    ValidationFailure,
    ValidationSuccess,
    ImageAPIRequest,
    make_request,
    make_random_string,
)


class ErrorOnRandomRegion(ValidationTest):
    name = "Random region gives 400"
    compliance_level = ComplianceLevel.LEVEL_1
    category = TestCategory.REGION
    versions = [IIIFVersion.V2, IIIFVersion.V3]

    @staticmethod
    def run(server: TargetServer) -> ValidationFailure | ValidationSuccess:
        url = ImageAPIRequest.of(region=make_random_string(6)).url(server)
        resp = make_request(url)
        if resp.status != 400:
            return ValidationFailure(
                url=url,
                expected="400 Bad Request",
                received=f"{resp.status} {resp.body.decode('utf8')}",
                details=f"Server did not return 400 for random region",
            )
        return ValidationSuccess(details="Server returned 400 for random region")



from .test import (
    ValidationTest,
    ComplianceLevel,
    TestCategory,
    IIIFVersion,
    TargetServer,
    ValidationFailure,
    ValidationSuccess,
    make_request,
    ImageAPIRequest,
    parse_rel_links,
)


class ImagesHaveCanonicalLinkHeader(ValidationTest):
    name = "Images have canonical Link header"
    compliance_level = ComplianceLevel.OPTIONAL
    category = TestCategory.HTTP
    versions = [IIIFVersion.V2, IIIFVersion.V3]
    extra_name = "canonicalLinkHeader"

    @staticmethod
    def run(server: TargetServer) -> ValidationFailure | ValidationSuccess:
        req = ImageAPIRequest.of()
        url = req.url(server)
        img_resp = make_request(url)
        if img_resp.status != 200:
            return ValidationFailure(
                url=url,
                expected="HTTP Status 200",
                received=f"HTTP Status {img_resp.status}",
                details="Got unexpected status code when requesting image",
            )
        if "link" not in img_resp.headers:
            return ValidationFailure(
                url=url,
                expected='Link header with rel="canonical"',
                received="No Link header",
                details="No Link header present in image response",
            )
        link_header = img_resp.headers["link"]
        rel_links = parse_rel_links(link_header)
        if "canonical" not in rel_links:
            return ValidationFailure(
                url=url,
                expected='Link header with rel="canonical"',
                received=f'Link header without rel="canonical": {link_header}',
                details="No canonical link found in Link header",
            )
        return ValidationSuccess(details="Image response had canonical Link header")



import json
from typing import Callable

from .test import (
    ValidationTest,
    ComplianceLevel,
    TestCategory,
    IIIFVersion,
    TargetServer,
    ValidationSuccess,
    ValidationFailure,
    get_info_json,
    make_request,
)


class TestInfoJsonStructure(ValidationTest):
    name = "Check info.json"
    compliance_level = ComplianceLevel.LEVEL_0
    category = TestCategory.INFO
    versions = [IIIFVersion.V2, IIIFVersion.V3]

    @staticmethod
    def run(server: TargetServer) -> list[ValidationFailure | ValidationSuccess]:
        results = []
        info_resp = make_request(server.info_url())
        if info_resp.status != 200:
            url = server.info_url()
            return [
                ValidationFailure(
                    url=url,
                    expected="status code 200",
                    received=f"status code {info_resp.status}",
                    details="Failed to fetch info.json",
                )
            ]

        content_type = info_resp.headers.get("content-type", "").split(";")[0]
        if content_type not in ("application/json", "application/ld+json"):
            results.append(
                ValidationFailure(
                    url=server.info_url(),
                    expected="content-type is application/json or application/ld+json",
                    received=f"content-type is {content_type}",
                    details="Invalid content-type for info.json response",
                )
            )
        try:
            json.loads(info_resp.body)
        except ValueError:
            results.append(
                ValidationFailure(
                    url=server.info_url(),
                    expected="valid JSON",
                    received="invalid JSON",
                    details="info.json response must be valid JSON",
                )
            )
            return results

        info = get_info_json(server)
        is_v2 = IIIFVersion.V2 == server.version
        id_field = "@id" if is_v2 else "id"
        context = f"http://iiif.io/api/image/{2 if is_v2 else 3}/context.json"
        info_url = server.info_url()
        base_url = f"{server.base_url}/{server.validation_id}"
        results.extend(
            [
                has_required_field(info, "width", url=info_url),
                field_has_type(info, "width", int, url=info_url),
                has_required_field(info, "height", url=info_url),
                field_has_type(info, "height", int, url=info_url),
                has_required_field(info, id_field, url=info_url),
                field_has_value_matching(
                    info,
                    id_field,
                    lambda val: isinstance(val, str) and val.startswith("http"),
                    "is an URL",
                    url=info_url,
                ),
                field_has_value(info, id_field, base_url, url=info_url),
                has_required_field(info, "@context", url=info_url),
                field_has_value(info, "@context", context, url=info_url),
                has_required_field(info, "protocol", url=info_url),
                field_has_value(
                    info, "protocol", "http://iiif.io/api/image", url=info_url
                ),
                *check_profile(is_v2, info, url=info_url),
                *check_sizes(info, url=info_url),
                *check_tiles(info, url=info_url),
                *(check_v3_fields(info, url=info_url) if not is_v2 else []),
            ]
        )
        return results


def has_required_field(
    info: dict, field: str, url: str | None = None
) -> ValidationFailure | ValidationSuccess:
    return check(f"required field: {field}", field in info, url=url)


def field_has_type(
    info: dict, field: str, type_: type, url: str | None = None
) -> ValidationFailure | ValidationSuccess:
    return check(
        f"is {type_} field: {field}", isinstance(info.get(field), type_), url=url
    )


def field_has_value(
    info: dict, field: str, value: object, url: str | None = None
) -> ValidationFailure | ValidationSuccess:
    return check(
        f"field {field} has value: {value}",
        info.get(field) == value,
        str(value),
        str(info.get(field)),
        url=url,
    )


def field_has_value_matching(
    info: dict,
    field: str,
    predicate: Callable,
    description: str,
    url: str | None = None,
) -> ValidationFailure | ValidationSuccess:
    return check(
        f"field {field} {description}",
        predicate(info.get(field)),
        description,
        str(info.get(field)),
        url=url,
    )


def check_profile(
    is_v2: bool, info: dict, url: str | None = None
) -> list[ValidationFailure | ValidationSuccess]:
    results = []
    results.append(has_required_field(info, "profile", url=url))
    if is_v2:
        results.extend(
            [
                field_has_type(info, "profile", list, url=url),
                field_has_value_matching(
                    info,
                    "profile",
                    lambda p: isinstance(p, list)
                    and p[0].startswith("http://iiif.io/api/image/2/level"),
                    "should have an official IIIF profile as the first profile",
                    url=url,
                ),
            ]
        )
    else:
        results.append(
            field_has_value_matching(
                info,
                "profile",
                lambda p: p in ["level0", "level1", "level2"],
                "should be one of ['level0', 'level1', 'level2']",
                url=url,
            )
        )
    return results


def check_sizes(
    info: dict, url: str | None = None
) -> list[ValidationFailure | ValidationSuccess]:
    sizes = info.get("sizes")
    if sizes is None:
        return []

    results = [field_has_type(info, "sizes", list, url=url)]
    if not isinstance(sizes, list):
        return results

    for size in sizes:
        if not isinstance(size, dict):
            results.append(
                ValidationFailure(
                    url=url or "<unknown>",
                    expected="dict/object",
                    received=str(type(size)),
                    details="entries in 'sizes' must be objects",
                )
            )
        results.extend(
            [
                has_required_field(size, "width", url=url),
                field_has_type(size, "width", int, url=url),
                has_required_field(size, "height", url=url),
                field_has_type(size, "height", int, url=url),
            ]
        )
    return results


def check_tiles(
    info: dict, url: str | None = None
) -> list[ValidationFailure | ValidationSuccess]:
    tiles = info.get("tiles")
    if tiles is None:
        return []

    results = [field_has_type(info, "tiles", list, url=url)]
    if not isinstance(tiles, list):
        return results

    for tile in tiles:
        if not isinstance(tile, dict):
            results.append(
                ValidationFailure(
                    url=url or "<unknown>",
                    expected="dict/object",
                    received=str(type(tile)),
                    details="entries in 'tiles' must be objects",
                )
            )
        results.extend(
            [
                has_required_field(tile, "width", url=url),
                field_has_type(tile, "width", int, url=url),
                has_required_field(tile, "scaleFactors", url=url),
                field_has_value_matching(
                    tile,
                    "scaleFactors",
                    lambda f: isinstance(f, list)
                    and all(isinstance(x, int) for x in f),
                    "scaleFactors must be integer values",
                    url=url,
                ),
            ]
        )
    return results


def check_v3_fields(
    info: dict, url: str | None = None
) -> list[ValidationFailure | ValidationSuccess]:
    results = []
    results.append(has_required_field(info, "type", url=url))
    results.append(field_has_value(info, "type", "ImageService3", url=url))
    if "license" in info:
        results.append(
            ValidationFailure(
                url=url or "<unknown>",
                expected="field name 'rights'",
                received="field name 'license'",
                details="'license' has been renamed to 'rights' in v3",
            )
        )
    if "rights" in info:
        results.append(field_has_type(info, "rights", str, url=url))
        results.append(
            field_has_value_matching(
                info,
                "rights",
                lambda r: isinstance(r, str) and r.startswith("http:"),
                "'rights' must be a single URI from Creative Commonsl, RightsStatements.org or URIs registered as extensions",
                url=url,
            )
        )
    if "extraQualities" in info:
        field_has_type(info, "extraQualities", list, url=url)
        field_has_value_matching(
            info,
            "extraQualities",
            lambda qs: isinstance(qs, list) and all(isinstance(q, str) for q in qs),
            "'extraQualities' must be a list of strings",
            url=url,
        )
    if "extraFormats" in info:
        field_has_type(info, "extraFormats", list, url=url)
        field_has_value_matching(
            info,
            "extraFormats",
            lambda qs: isinstance(qs, list) and all(isinstance(q, str) for q in qs),
            "'extraFormats' must be a list of strings",
            url=url,
        )
    if "extraFeatures" in info:
        field_has_type(info, "extraFeatures", list, url=url)
        field_has_value_matching(
            info,
            "extraFeatures",
            lambda qs: isinstance(qs, list) and all(isinstance(q, str) for q in qs),
            "'extraFeatures' must be a list of strings",
            url=url,
        )
    results.extend(check_linking_prop(info, "service", url=url))
    results.extend(check_linking_prop(info, "partOf", url=url))
    results.extend(check_linking_prop(info, "seeAlso", url=url))

    if "attribution" in info:
        results.append(
            ValidationFailure(
                url=url or "<unknown>",
                expected="'attribution' field not to be present",
                received="'attribution' field was present",
                details="'attribution' field was removed in v3",
            )
        )
    if "logo" in info:
        results.append(
            ValidationFailure(
                url=url or "<unknown>",
                expected="'logo' field not to be present",
                received="'logo' field was present",
                details="'logo' field was removed in v3",
            )
        )
    return results
    return results


def check_linking_prop(
    info: dict, field: str, url: str | None = None
) -> list[ValidationFailure | ValidationSuccess]:
    if field not in info:
        return []
    results = []
    results.append(field_has_type(info, field, list, url=url))
    if not isinstance(info[field], list):
        return results

    for item in info[field]:
        if not isinstance(item, dict):
            results.append(
                ValidationFailure(
                    url=url or "<unknown>",
                    expected="type 'list'",
                    received=f"type '{type(item)}",
                    details=f"linking entries in {field} must be dicts/objects",
                )
            )
        else:
            results.append(
                ValidationSuccess(
                    details=f"Linking entries in {field} are dicts/objects"
                )
            )
        results.extend(
            [
                has_required_field(item, "id", url=url),
                field_has_type(item, "id", str, url=url),
                has_required_field(item, "type", url=url),
                field_has_type(item, "type", str, url=url),
            ]
        )

    return results


def check_label(obj: dict) -> list[ValidationFailure | ValidationSuccess]:
    raise NotImplementedError


def check(
    description: str,
    condition: bool,
    expected: str = "👍",
    received: str = "👎",
    url: str | None = None,
) -> ValidationFailure | ValidationSuccess:
    if condition:
        return ValidationSuccess(details=f"info.json passed: {description}")
    else:
        return ValidationFailure(
            url=url or "<unknown>",
            expected=expected,
            received=received,
            details=f"info.json failed: {description}",
        )



from .test import (
    ValidationTest,
    ComplianceLevel,
    TestCategory,
    IIIFVersion,
    TargetServer,
    ValidationFailure,
    ValidationSuccess,
    ImageAPIRequest,
    get_image,
    is_bitonal,
)


class BitonalImagesAreReturned(ValidationTest):
    name = "Bitonal image is returned when requesting quality=bitonal"
    compliance_level = ComplianceLevel.OPTIONAL
    category = TestCategory.QUALITY
    versions = [IIIFVersion.V2, IIIFVersion.V3]
    extra_name = "bitonal"

    @staticmethod
    def run(server: TargetServer) -> ValidationFailure | ValidationSuccess:
        req = ImageAPIRequest.of(quality="bitonal")
        img = get_image(server, req)
        if not is_bitonal(img):
            url = req.url(server)
            return ValidationFailure(
                url=url,
                expected="Bitonal image",
                received="Non-bitonal image",
                details="Server did not return a bitonal image when requesting quality=bitonal",
            )
        return ValidationSuccess(
            details="Server returned a bitonal image when requesting quality=bitonal"
        )



from .test import (
    ValidationTest,
    ComplianceLevel,
    TestCategory,
    IIIFVersion,
    TargetServer,
    ValidationFailure,
    ValidationSuccess,
    ImageAPIRequest,
    get_image,
    get_expected_image,
    compare_images,
)


class RotationMirror180Test(ValidationTest):
    name = "Mirroring plus 180 rotation"
    compliance_level = ComplianceLevel.OPTIONAL
    category = TestCategory.ROTATION
    versions = [IIIFVersion.V2, IIIFVersion.V3]
    extra_name = "mirroring"

    @staticmethod
    def run(server: TargetServer) -> ValidationSuccess | ValidationFailure:
        request = ImageAPIRequest.of(rotation="!180")
        try:
            transformed_img = get_image(server, request)
            expected_img = get_expected_image()

            if (
                transformed_img.width != expected_img.width
                or transformed_img.height != expected_img.height
            ):
                url = request.url(server)
                return ValidationFailure(
                    url=url,
                    expected=f"Image dimensions to be ({expected_img.width}, {expected_img.height})",
                    received=f"Image dimensions were ({transformed_img.width}, {transformed_img.height})",
                    details="Incorrect image dimensions for mirrored 180 degree rotation.",
                )

            # !180 is a vertical flip
            expected_transformed_img = expected_img.flip("vertical")

            if compare_images(transformed_img, expected_transformed_img):
                return ValidationSuccess(
                    details="Mirroring plus 180 degree rotation was successful."
                )
            else:
                url = request.url(server)
                return ValidationFailure(
                    url=url,
                    expected="A correctly transformed image (mirrored and rotated 180 degrees).",
                    received="An incorrectly transformed image.",
                    details="Image does not match the expected transformation.",
                )
        except Exception as e:
            url = request.url(server)
            return ValidationFailure(
                url=url,
                expected="A valid image.",
                received=f"An error: {e}",
                details="Request for mirrored 180 degree rotation failed.",
            )



import random

from .test import (
    ValidationTest,
    ComplianceLevel,
    TestCategory,
    IIIFVersion,
    TargetServer,
    ValidationFailure,
    ValidationSuccess,
    ImageAPIRequest,
    make_request,
)


class SizeNoUpTest(ValidationTest):
    name = "Size greater than 100% should only work with the ^ notation"
    compliance_level = ComplianceLevel.LEVEL_1
    category = TestCategory.SIZE
    versions = [IIIFVersion.V3]

    @staticmethod
    def run(server: TargetServer) -> list[ValidationSuccess | ValidationFailure]:
        random.seed(31337)
        s = random.randint(1100, 2000)
        sizes_to_check = [f"{s},{s}", f",{s}", f"{s},", "pct:200", "!2000,3000"]

        results = []
        for size_str in sizes_to_check:
            req = ImageAPIRequest.of(size=size_str)
            url = req.url(server)
            resp = make_request(url)

            if resp.status != 200:
                results.append(
                    ValidationSuccess(
                        details=f"Upscaling size='{size_str}' failed as expected."
                    )
                )
            else:
                results.append(
                    ValidationFailure(
                        url=url,
                        expected="Non-200 status code",
                        received=f"{resp.status}",
                        details=f"Retrieving upscaled image with size='{size_str}' succeeded but should have failed.",
                    )
                )
        return results



from .test import (
    ValidationTest,
    ComplianceLevel,
    TestCategory,
    IIIFVersion,
    TargetServer,
    ValidationFailure,
    ValidationSuccess,
    ImageAPIRequest,
    get_image,
    is_grayscale,
)


class GrayscaleImagesAreReturned(ValidationTest):
    name = "Grayscale image is returned when requesting quality=gray"
    compliance_level = ComplianceLevel.LEVEL_2
    category = TestCategory.QUALITY
    versions = [IIIFVersion.V2, IIIFVersion.V3]
    extra_name = "gray"

    @staticmethod
    def run(server: TargetServer) -> ValidationFailure | ValidationSuccess:
        req = ImageAPIRequest.of(quality="gray")
        img = get_image(server, req)
        if not is_grayscale(img):
            url = req.url(server)
            return ValidationFailure(
                url=url,
                expected="Grayscale image",
                received=f"Non-grayscale image: {img.interpretation}",
                details="Server did not return a grayscale image when requesting quality=gray",
            )
        return ValidationSuccess(
            details="Server returned a grayscale image when requesting quality=gray"
        )



from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Iterable, Iterator
from urllib.error import HTTPError

from .tests.test import (
    get_tests,
    ServerFeatures,
    TargetServer,
    TestCategory,
    ValidationFailure,
    ValidationSuccess,
    ValidationTest,
)
from .tests.http_caching import HttpCachingTest


class SizedGenerator[T]:
    def __init__(self, it: Iterable[T], total: int):
        self._iterator = iter(it)
        self.num_total = total

    def __iter__(self) -> Iterator[T]:
        return self._iterator

    def __next__(self) -> T:
        return next(self._iterator)


def _test_is_applicable(test, features: ServerFeatures):
    if test == HttpCachingTest and not features.supports_caching:
        return False

    if features.version not in test.versions:
        return False
    level = test.compliance_level
    if isinstance(level, dict):
        level = level[features.version]
    if level <= features.compliance_level:
        return True
    extra_names = test.extra_name
    if extra_names is None:
        return False
    elif isinstance(extra_names, str):
        extra_names = [extra_names]
    if test.category == TestCategory.FORMAT and any(
        name in features.extra_formats for name in extra_names
    ):
        return True
    if test.category == TestCategory.QUALITY and any(
        name in features.extra_qualities for name in extra_names
    ):
        return True
    if any(name in features.extra_features for name in extra_names):
        return True
    return False


def run_tests(
    server: TargetServer, max_threads=4
) -> SizedGenerator[
    tuple[
        type[ValidationTest],
        ValidationFailure
        | ValidationSuccess
        | list[ValidationFailure | ValidationSuccess],
    ]
]:
    features = server.feature_set()
    all_tests = get_tests(server.version)

    applicable_tests = [
        test for test in all_tests if _test_is_applicable(test, features)
    ]

    with ThreadPoolExecutor(max_workers=max_threads) as executor:
        future_to_test = {
            executor.submit(test.run, server): test for test in applicable_tests
        }

        def _generator():
            for future in as_completed(future_to_test):
                test = future_to_test[future]
                try:
                    result = future.result()
                except Exception as exc:
                   result = ValidationFailure(
                       url= exc.url if isinstance(exc, HTTPError) else "<unknown>",
                       expected="Test to complete without error",
                       received=f"Test raised an exception: {exc}",
                       details="An unexpected error occurred during the test execution",
                   )
                yield test, result

        return SizedGenerator(_generator(), len(applicable_tests))



from .test import (
    ValidationTest,
    ComplianceLevel,
    TestCategory,
    IIIFVersion,
    TargetServer,
    ValidationFailure,
    ValidationSuccess,
    get_image,
    ImageAPIRequest,
    get_expected_image,
    compare_images,
)


class FullImageMatches(ValidationTest):
    name = "Correctly serves full image"
    compliance_level = ComplianceLevel.LEVEL_0
    category = TestCategory.INFO
    versions = [IIIFVersion.V2, IIIFVersion.V3]

    @staticmethod
    def run(server: TargetServer) -> ValidationFailure | ValidationSuccess:
        req = ImageAPIRequest.of()
        given_image = get_image(server, req)
        expected_image = get_expected_image()
        if compare_images(given_image, expected_image):
            return ValidationSuccess(details="Full image matches expected image")
        else:
            url = req.url(server)
            return ValidationFailure(
                url=url,
                expected="Full image to match expected image",
                received="Full image did not match expected image",
                details="Full image did not match expected image",
            )



from .test import (
    ValidationTest,
    ComplianceLevel,
    TestCategory,
    IIIFVersion,
    TargetServer,
    ValidationFailure,
    ValidationSuccess,
    make_request,
)


class ErrorEscapedSlash(ValidationTest):
    name = "Forward slash gives 404"
    compliance_level = ComplianceLevel.LEVEL_1
    category = TestCategory.INFO
    versions = [IIIFVersion.V2, IIIFVersion.V3]

    @staticmethod
    def run(server: TargetServer) -> ValidationFailure | ValidationSuccess:
        resp = make_request(f"{server.base_url}/a/b/info.json")
        if resp.status == 404:
            return ValidationSuccess(
                details="Returned 404 for forward slash in identifier"
            )
        else:
            url = f"{server.base_url}/a/b/info.json"
            return ValidationFailure(
                url=url,
                expected="HTTP Status 404",
                received=f"HTTP Status {resp.status}",
                details="Got unexpected status code for identifier with forward slash",
            )



from urllib.error import HTTPError

from .test import (
    ValidationTest,
    ComplianceLevel,
    TestCategory,
    IIIFVersion,
    TargetServer,
    ValidationFailure,
    ValidationSuccess,
    ImageAPIRequest,
    make_random_string,
    make_request,
)


class RotationErrorRandomTest(ValidationTest):
    name = "Random rotation gives 400"
    compliance_level = ComplianceLevel.LEVEL_1
    category = TestCategory.ROTATION
    versions = [IIIFVersion.V2, IIIFVersion.V3]

    @staticmethod
    def run(server: TargetServer) -> ValidationSuccess | ValidationFailure:
        random_string = make_random_string(4)
        request = ImageAPIRequest.of(rotation=random_string)
        url = request.url(server)
        resp = make_request(url)
        if resp.status != 400:
            return ValidationFailure(
                url=url,
                expected="HTTP Status 400",
                received=f"HTTP Status {resp.status}",
                details=f"Server did not return an error for invalid rotation value '{random_string}'.",
            )
        else:
            return ValidationSuccess(
                details=f"Server correctly returned HTTP 400 for invalid rotation value '{random_string}'."
            )



import random

from .test import (
    ValidationTest,
    ComplianceLevel,
    TestCategory,
    IIIFVersion,
    TargetServer,
    ValidationFailure,
    ValidationSuccess,
    ImageAPIRequest,
    get_expected_image,
    get_image,
    compare_rectangle,
)


class RegionPixels(ValidationTest):
    name = "Region specified by pixels"
    compliance_level = ComplianceLevel.LEVEL_1
    category = TestCategory.REGION
    versions = [IIIFVersion.V2, IIIFVersion.V3]
    extra_name = "regionByPx"

    @staticmethod
    def run(server: TargetServer) -> list[ValidationFailure | ValidationSuccess]:
        random.seed(31337)  # for reproducibility
        expected = get_expected_image()
        results = []
        for _ in range(5):
            x = random.randint(0, 9)
            y = random.randint(0, 9)
            region = f"{x * 100},{y * 100},100,100"
            req = ImageAPIRequest.of(region=region)
            rect = get_image(server, req)
            if compare_rectangle(expected, rect, x, y):
                results.append(ValidationSuccess(details=f"Region {region} matches"))
            else:
                results.append(
                    ValidationFailure(
                        url=req.url(server),
                        expected=f"Region {region} to match expected image region",
                        received="Region does not match expected image region",
                        details=f"Region {region} does not match",
                    )
                )
        return results



from .test import (
    ValidationTest,
    ComplianceLevel,
    TestCategory,
    IIIFVersion,
    TargetServer,
    ValidationFailure,
    ValidationSuccess,
    ImageAPIRequest,
    make_request,
)


class SizeNoFullTest(ValidationTest):
    name = "Version 3.0 has replaced the size full with max"
    compliance_level = ComplianceLevel.LEVEL_0
    category = TestCategory.SIZE
    versions = [IIIFVersion.V3]

    @staticmethod
    def run(server: TargetServer) -> ValidationSuccess | ValidationFailure:
        req = ImageAPIRequest.of(size="full")
        url = req.url(server)
        resp = make_request(url)

        if resp.status != 200:
            return ValidationSuccess(
                details="Request for size='full' did not return 200, as expected in v3."
            )
        else:
            return ValidationFailure(
                url=url,
                expected="Non-200 status code",
                received=f"{resp.status}",
                details="Version 3.0 has replaced the size 'full' with 'max'. A request for 'full' should not return a 200 status.",
            )



from __future__ import annotations

import inspect
import importlib.resources
import json
import pkgutil
import random
import re
from collections import defaultdict
from enum import Enum
from urllib.request import urlopen, Request, OpenerDirector
from urllib.error import HTTPError
from typing import ClassVar, NamedTuple, Protocol, runtime_checkable

from pyvips import Image

#: Standard headers to use for requests, based on those
# used by the official validator.
HEADERS = {
    "Origin": "http://iiif.io/",
    "Referer": "http://iiif.io/api/image/validator",
    "User-Agent": "Mozilla/5.0 (Macintosh; U; Intel Mac OS X 10.5; en-US; rv:1.9.1b3pre) Gecko/20081130 Minefield/3.1b3pre",
}
VIPS_FORMAT_BITS = {
    "uchar": 8,
    "char": 8,
    "ushort": 16,
    "short": 16,
    "uint": 32,
    "int": 32,
    "float": 32,
    "complex": 64,
    "double": 64,
    "dpcomplex": 128,
}


class TestCategory(str, Enum):
    INFO = "Info/Identifier"
    REGION = "Region"
    SIZE = "Size"
    ROTATION = "Rotation"
    QUALITY = "Quality"
    FORMAT = "Format"
    HTTP = "HTTP"


class ComplianceLevel(int, Enum):
    LEVEL_0 = 0
    LEVEL_1 = 1
    LEVEL_2 = 2
    OPTIONAL = 3

    @classmethod
    def from_profile(cls, profile: str | list[str | dict]) -> ComplianceLevel:
        if isinstance(profile, list):
            if not isinstance(profile[0], str):
                raise ValueError(
                    f"Invalid profile in info.json, first profile in a v2 `profile` array must be a string with an IIIF compliance level, was: {profile[0]!r}"
                )
            profile = profile[0]
        if profile == "level2" or profile == "http://iiif.io/api/image/2/level2.json":
            return cls.LEVEL_2
        elif profile == "level1" or profile == "http://iiif.io/api/image/2/level1.json":
            return cls.LEVEL_1
        elif profile == "level0" or profile == "http://iiif.io/api/image/2/level0.json":
            return cls.LEVEL_0
        raise ValueError(f"Unknown IIIF compliance level in profile: {profile!r}")


class IIIFVersion(str, Enum):
    V2 = "2.0"
    V3 = "3.0"


class ValidationSuccess(NamedTuple):
    success: bool = True
    details: str = "Validation successful"


class ValidationFailure(NamedTuple):
    expected: str
    received: str
    url: str
    success: bool = False
    details: str = "Validation failed"


@runtime_checkable
class ValidationTest(Protocol):
    name: ClassVar[str]
    category: ClassVar[TestCategory]
    compliance_level: ClassVar[ComplianceLevel | dict[IIIFVersion, ComplianceLevel]]
    versions: ClassVar[list[IIIFVersion]]
    extra_name: ClassVar[str | list[str] | None] = None

    @staticmethod
    def run(
        server: TargetServer,
    ) -> (
        ValidationSuccess
        | ValidationFailure
        | list[ValidationSuccess | ValidationFailure]
    ): ...


class TargetServer(NamedTuple):
    base_url: str
    validation_id: str
    version: IIIFVersion

    def info_url(self) -> str:
        return f"{self.base_url}/{self.validation_id}/info.json"

    def feature_set(self) -> ServerFeatures:
        resp = make_request(self.info_url())
        supports_caching = (
            "etag" in resp.headers or "last-modified" in resp.headers
        )
        return ServerFeatures.from_info_json(json.loads(resp.body), supports_caching)



class ServerFeatures(NamedTuple):
    version: IIIFVersion
    compliance_level: ComplianceLevel
    extra_qualities: list[str]
    extra_formats: list[str]
    extra_features: list[str]
    supports_caching: bool

    @classmethod
    def from_info_json(cls, info: dict, supports_caching: bool) -> ServerFeatures:
        compliance = ComplianceLevel.from_profile(info["profile"])
        if "id" in info:
            version = IIIFVersion.V3
            extra_qualities = info.get("extraQualities", [])
            extra_formats = info.get("extraFormats", [])
            extra_features = info.get("extraFeatures", [])
        else:
            version = IIIFVersion.V2
            profile = info["profile"]
            extra_qualities = []
            extra_formats = []
            extra_features = []
            if isinstance(profile, list) and len(profile) > 1:
                for p in profile[1:]:
                    if isinstance(p, dict) and "qualities" in p:
                        extra_qualities.extend([q for q in p["qualities"]])
                    if isinstance(p, dict) and "formats" in p:
                        extra_formats.extend([f for f in p["formats"]])
                    if isinstance(p, dict) and "supports" in p:
                        extra_features.extend(p["supports"])
        return cls(
            version=version,
            compliance_level=compliance,
            extra_qualities=extra_qualities,
            extra_formats=extra_formats,
            extra_features=extra_features,
            supports_caching=supports_caching,
        )


class HttpResponse(NamedTuple):
    status: int
    body: bytes
    headers: dict[str, str]


class ImageAPIRequest(NamedTuple):
    region: str
    size: str
    rotation: str
    quality: str
    format: str

    @classmethod
    def of(
        cls,
        region: str = "full",
        size: str = "max",
        rotation: str = "0",
        quality: str = "default",
        format: str = "jpg",
    ) -> "ImageAPIRequest":
        return cls(
            region=region,
            size=size,
            rotation=rotation,
            quality=quality,
            format=format,
        )

    def to_v2(self) -> "ImageAPIRequest":
        size = self.size
        if size == "max":
            size = "full"
        return ImageAPIRequest(
            region=self.region,
            size=size,
            rotation=self.rotation,
            quality=self.quality,
            format=self.format,
        )

    def url(self, server: TargetServer) -> str:
        if server.version == IIIFVersion.V2:
            req = self.to_v2()
        else:
            req = self
        return f"{server.base_url}/{server.validation_id}/{req.region}/{req.size}/{req.rotation}/{req.quality}.{req.format}"


def make_request(
    req_or_url: str | Request,
    extra_headers: dict[str, str] = {},
    opener: OpenerDirector | None = None,
) -> HttpResponse:
    all_headers = {**HEADERS, **extra_headers}
    if isinstance(req_or_url, str):
        req = Request(req_or_url, headers=all_headers)
    else:
        req = req_or_url
        for k, v in all_headers.items():
            req.add_header(k, v)
    try:
        if opener:
            resp = opener.open(req)
        else:
            resp = urlopen(req)
        return HttpResponse(
            resp.status, resp.read(), {k.lower(): v for k, v in resp.headers.items()}
        )
    except HTTPError as e:
        return HttpResponse(
            e.code, e.read(), {k.lower(): v for k, v in e.headers.items()}
        )


def get_info_json(server: TargetServer) -> dict:
    url = f"{server.base_url}/{server.validation_id}/info.json"
    response = make_request(url)
    if response.status != 200:
        raise Exception(f"Failed to fetch info.json: {response.status}")
    return json.loads(response.body)


def get_image(server: TargetServer, request: ImageAPIRequest) -> Image:
    if server.version == IIIFVersion.V2:
        request = request.to_v2()

    url = f"{server.base_url}/{server.validation_id}/{request.region}/{request.size}/{request.rotation}/{request.quality}.{request.format}"
    response = make_request(url)
    if response.status != 200:
        raise Exception(
            f"Failed to fetch image: {response.status} {response.body.decode('utf-8', errors='replace')}"
        )
    return Image.new_from_buffer(response.body, "")


def get_expected_image() -> Image:
    img_path = importlib.resources.files("iiif_validator") / "validation_image.png"
    return Image.new_from_file(str(img_path))


def compare_rectangle(reference: Image, rect: Image, x: int, y: int) -> bool:
    if rect.width != 100 or rect.height != 100:
        return False
    # Omit 13 pixel border and compare a 74x74 square in the center
    square_reference = reference.extract_area(x * 100 + 13, y * 100 + 13, 74, 74)
    square_rect = rect.extract_area(13, 13, 74, 74)
    diff = (square_reference - square_rect).abs().max()
    # Allow minor differences due to compression
    return diff < 6


def compare_images(
    given: Image, expected: Image, x: int | None = None, y: int | None = None
) -> bool:
    # For similarity, subtract the two images and get the average color value
    # since the reference image has no blacks, the closer we are to black
    # the better the match. We'll never get full black due to compression
    # artifacts, but 6 is a good threshold
    if x is not None and y is not None:
        # Omit 13 pixel border and compare a 74x74 square in the center
        square_given = given.extract_area(x * 100 + 13, y * 100 + 13, 74, 74)
        square_expected = expected.extract_area(x * 100 + 13, y * 100 + 13, 74, 74)
        diff = (square_given - square_expected).abs().avg()
    else:
        diff = (given - expected).abs().avg()
    # Allow minor differences due to compression
    return diff < 6


def get_image_format(server: TargetServer, request: ImageAPIRequest) -> str:
    img = get_image(server, request)
    loader = img.get("vips-loader")
    return loader[: loader.index("load")]


def make_random_string(length: int) -> str:
    return (
        "".join(chr(random.randint(48, 122)) for _ in range(length))
        .replace("?", "$")
        .replace("#", "$")
        .replace("/", "$")
    )


def parse_rel_links(header: str) -> dict[str, list[str]]:
    links = defaultdict(lambda: [])
    for link in re.finditer(r'<([^>]+)>;\s*rel="([^"]+)"\s*([,]*)', header):
        uri = link.group(1)
        rel = link.group(2)
        links[rel].append(uri)
    return links


def is_bitonal(img: Image, min_pct: float = 0.99) -> bool:
    if img.hasalpha():
        img = img[:-1]

    black = [0] * img.bands
    white = [(1 << VIPS_FORMAT_BITS[img.format]) - 1] * img.bands

    # Mask of pixels that are either black or white, values
    # will be 0 (false) or 255 (true)
    mask = ((img == black) | (img == white)).bandand()
    return mask.avg() / 255


def is_grayscale(img: Image, min_pct: float = 0.99) -> bool:
    if img.hasalpha():
        img = img[:-1]
    if img.bands == 1 or img.interpretation in ("grey16", "b-w"):
        return True

    # Mask of pixels where R == G == B, values will be 0 (false) or 255 (true)
    mask = (img[0] == img[1]) & (img[1] == img[2])
    return mask.avg() / 255 >= min_pct


def get_tests(iiif_version: IIIFVersion = IIIFVersion.V3) -> list[type[ValidationTest]]:
    assert __package__ is not None, "This function must be called from within a package"
    top_level_pkg = __package__.split(".")[0]
    tests_pkg_name = f"{top_level_pkg}.tests"

    tests_pkg = importlib.import_module(tests_pkg_name)

    implementations = []
    package_path = tests_pkg.__path__
    package_name = tests_pkg.__name__

    for module_info in pkgutil.walk_packages(
        path=package_path, prefix=f"{package_name}."
    ):
        if module_info.name == __name__:
            continue
        module = importlib.import_module(module_info.name)
        for _, member in inspect.getmembers(module, inspect.isclass):
            if member.__module__ == module.__name__ and (
                (ValidationTest in member.mro() or isinstance(member, ValidationTest))
                and iiif_version in member.versions
            ):
                implementations.append(member)

    return implementations



from .test import (
    ValidationTest,
    ComplianceLevel,
    TestCategory,
    IIIFVersion,
    TargetServer,
    ValidationFailure,
    ValidationSuccess,
    ImageAPIRequest,
    get_image_format,
)


class FormatJPEG(ValidationTest):
    name = "JPEG format"
    compliance_level = ComplianceLevel.LEVEL_0
    category = TestCategory.FORMAT
    versions = [IIIFVersion.V2, IIIFVersion.V3]

    @staticmethod
    def run(server: TargetServer) -> ValidationFailure | ValidationSuccess:
        req = ImageAPIRequest.of(format="jpg")
        fmt = get_image_format(server, req)
        if fmt == "jpeg":
            return ValidationSuccess(details="Server returned a JPEG image")
        else:
            url = req.url(server)
            return ValidationFailure(
                url=url,
                expected="JPEG image",
                received=f"{fmt or '<none>'} image",
                details="Server did not return a JPEG image",
            )
